# File: pca.py
# Author: Anne Sabourin
# Date: 2025

"""
Module implementing PCA for extremes
"""

import numpy as np
from scipy.linalg import svd

import matplotlib.pyplot as plt


class Xpca:
    """
    A class for performing Principal Component Analysis (PCA) on extreme points of a dataset.

    This class provides methods to fit the PCA model to training data, compute scores for test data,
    predict the projection of test data onto the principal axes, generate scree plots, and calculate
    reconstruction errors.

    Parameters:
    -----------
    thresh_train : float, optional
        Threshold for identifying extreme points during training.
    thresh_predict : float, optional
        Threshold for identifying extreme points during prediction.
    beta : float, default=1
        Scaling factor for rescaling extreme points.
    centered : bool, default=False
        Whether to center the data by subtracting the mean angle.

    Attributes:
    -----------
    thresh_train : float
        Threshold for identifying extreme points during training.
    thresh_predict : float
        Threshold for identifying extreme points during prediction.
    beta : float
        Scaling factor for rescaling extreme points.
    centered : bool
        Whether to center the data by subtracting the mean angle.
    mean_angle : array-like
        Mean angle of the extreme points.
    principal_comp : array-like
        Principal components of the extreme points.
    principal_axes : array-like
        Principal axes of the extreme points.
    eigenvalues : array-like
        Eigenvalues of the extreme points.
    """

    def __init__(self, thresh_train=None, thresh_predict=None, beta=1, centered=False):
        self.thresh_train = thresh_train
        self.thresh_predict = thresh_predict
        self.centered = centered
        self.beta = beta
        self.mean_angle = None
        self.principal_comp = None
        self.principal_axes = None
        self.eigenvalues = None
      
    def fit(self, X, max_dim=None):
        """
        Perform PCA by computing the singular value decomposition of the rescaled extreme points on the training dataset X.

        Each extreme point `x` is rescaled by the factor ||x||^{-beta}. The default value for beta is 1.
        Admissible values are in the interval (1 - alpha/2, 1], where alpha is the regular variation index of the input X.

        Parameters:
        -----------
        X : array-like
            Training data.
        max_dim : int, optional
            Maximum number of dimensions to consider for PCA.
        """
        Xt = X
        R_squared = np.sum(Xt**2, axis=1)
        norm_Xt = R_squared ** (1 / 2)
        # Set the training threshold if not provided
        if self.thresh_train is None:
            self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

        # set max_dim to ambient dimension if not provided
        if max_dim is None:
            max_dim = Xt.shape[1]
        # define extreme samples
        is_extreme = norm_Xt > self.thresh_train
        X_ext = Xt[is_extreme]
        # Update total mass accordingly
        self.number_extremes = len(X_ext)
        R_ext = norm_Xt[is_extreme]
        ang_ext = X_ext / (R_ext ** (self.beta)).reshape(-1, 1)
        if self.centered:
            center = np.mean(ang_ext, axis=0)
            self.mean_angle = center
            Z = ang_ext - center
        else:
            Z = ang_ext
        U, s, Vh = svd(Z, full_matrices=False)
        self.principal_comp = U[:, :max_dim]
        self.principal_axes = (Vh.transpose())[:, :max_dim]
        self.eigenvalues = s[:max_dim] ** 2

    def scores(self, Xtest, max_dim=None):
        """
        Return the scores, i.e., the coordinates of the points in the array Xtest,
        in the coordinate system given by the principal axes of the extreme points
        (with norm greater than self.thresh_predict) rescaled through the mapping
        x --> ||x||^{-beta} * x where beta = self.beta. The model must have been fit beforehand.
        If `Xtest` is the training data, this gives the usual PCA scores of the rescaled extreme data.

        Parameters:
        -----------
        Xtest : array-like
            Test data.
        max_dim : int, optional
            Maximum number of dimensions to consider for scores.

        Returns:
        --------
        coords_in_pr_axes : array-like
            Coordinates in the principal axes.
        X_ext : array-like
            Extreme points in the test data.
        mask : array-like
            Boolean mask indicating extreme points.
        """
        if self.principal_axes is None:
            raise ValueError("The model has not been fitted")
        if max_dim is None:
            p = np.shape(self.principal_axes)[1]
        else:
            p = max_dim

        if self.thresh_predict is None:
            self.thresh_predict = self.thresh_train

        Xt = Xtest
        norm_Xt = np.sum(Xt**2, axis=1) ** (1 / 2)
        # define extreme samples
        is_extreme = norm_Xt > self.thresh_predict
        X_ext = Xt[is_extreme]
        R_ext = norm_Xt[is_extreme]
        ang_ext = X_ext / (R_ext ** (self.beta)).reshape(-1, 1)

        if not self.centered:
            Zang = ang_ext
        else:
            Zang = ang_ext - self.mean_angle
        coords_in_pr_axes = Zang @ self.principal_axes[:, :p]
        mask = is_extreme
        return coords_in_pr_axes, X_ext, mask

    def predict(self, X_test, max_dim=None):
        """
        Project X_test onto the principal axes.

        Parameters:
        -----------
        X_test : array-like
            Test data.
        max_dim : int, optional
            Maximum number of dimensions to consider for prediction.

        Returns:
        --------
        recons_X : array-like
            Reconstructed data.
        X_ext : array-like
            Extreme points in the test data.
        mask : array-like
            Boolean mask indicating extreme points.
        """
        if self.principal_axes is None:
            raise ValueError("The model has not been fitted")
        if max_dim is None:
            p = np.shape(self.principal_axes)[1]
        else:
            p = max_dim
        Scores, X_ext, mask = self.scores(X_test, max_dim)
        recons_angle_intermediate = Scores @ ((self.principal_axes[:, :p]).transpose())
        if self.centered:
            recons_angle = recons_angle_intermediate + self.mean_angle
        else:
            recons_angle = recons_angle_intermediate
        R_ext = (np.sum(X_ext**2, axis=1)) ** (1 / 2)
        recons_X = recons_angle * (R_ext**self.beta).reshape(-1, 1)
        return recons_X, X_ext, mask

    def screeplot(self):
        """
        Generate a scree plot of the eigenvalues.

        Raises:
        -------
        ValueError
            If the model has not been fitted.
        """
        if self.principal_axes is None:
            raise ValueError("The model has not been fitted")
        eigenvals = self.eigenvalues
        # Create the scree plot
        plt.figure(figsize=(10, 6))
        plt.plot(eigenvals, "bo-", markersize=8, label="Eigenvalues")
        # Add labels and title
        plt.xlabel("Principal Component Index")
        plt.ylabel("Eigenvalue")
        plt.title("Scree Plot")
        plt.legend()
        plt.grid(True)
        plt.show()

    def recons_error(self, X_test, max_dim, test_beta=None):
        """
        Calculate the reconstruction error for the test data.

        The test data are rescaled to ensure existence of second moments, namely each test data is transformed following
        z = ||xtest||^{-beta} xtest, where beta <= 1 is given by the argument `test_beta`.
        Default is `test_beta=None`, in which case `test_beta` is set to the training value self.beta which defaults to 1.

        Parameters:
        -----------
        X_test : array-like
            Test data.
        max_dim : int
            Maximum number of dimensions to consider for reconstruction.
        test_beta : float, optional
            Scaling factor for rescaling test data.

        Returns:
        --------
        Err : array-like
            Reconstruction error.
        """
        if test_beta is None:
            beta = self.beta
        else:
            beta = test_beta
        recons_X, X_ext, mask = self.predict(X_test, max_dim)

        X_ext_norm = np.sum(X_ext**2, axis=1) ** (1 / 2)
        Z = X_ext / (X_ext_norm**beta).reshape(-1, 1)
        recons_Z = recons_X / (X_ext_norm**beta).reshape(-1, 1)
        Err = np.sum((Z - recons_Z) ** 2, axis=1)
        return Err


# """
# File: pca.py
# Author: Anne Sabourin
# Date: 2025-05-04
# """

# import numpy as np
# from scipy.linalg import svd

# # from math import isclose
# import matplotlib.pyplot as plt
# # from copy import deepcopy
# # from ...utils.EVT_basics import rank_transform, round_signif


# class Xpca:
#     def __init__(self, thresh_train=None, thresh_predict=None, beta=1, centered=False):
#         self.thresh_train = thresh_train
#         self.thresh_predict = thresh_predict
#         self.centered = centered
#         self.beta = beta
#         self.mean_angle = None
#         self.principal_comp = None
#         self.principal_axes = None
#         self.eigenvalues = None

#     def fit(self, X, max_dim=None):
#         """
#         Performs PCA by computing the singular value decomposition
#         of the rescaled extreme points on the training dataset X,
#         where each extreme point `x` is rescaled by the factor
#         || x ||^{-beta}. Default value for beta is
#         1. Admissible values are the interval  (1- alpha/2, 1]
#         where alpha  is the regular variation index of the input
#         X.

#         """
#         Xt = X
#         R_squared = np.sum(Xt**2, axis=1)
#         norm_Xt = R_squared ** (1 / 2)
#         # Set the training threshold if not provided
#         if self.thresh_train is None:
#             self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

#         # set max_dim to ambient dimension if not provided
#         if max_dim is None:
#             max_dim = Xt.shape[1]
#         # define extreme samples
#         is_extreme = norm_Xt > self.thresh_train
#         X_ext = Xt[is_extreme]
#         # Update total mass accordingly
#         self.number_extremes = len(X_ext)
#         R_ext = norm_Xt[is_extreme]
#         ang_ext = X_ext / (R_ext ** (self.beta)).reshape(-1, 1)
#         if self.centered:
#             center = np.mean(ang_ext, axis=0)
#             self.mean_angle = center
#             Z = ang_ext - center
#         else:
#             Z = ang_ext
#         U, s, Vh = svd(Z, full_matrices=False)
#         self.principal_comp = U[:, :max_dim]
#         self.principal_axes = (Vh.transpose())[:, :max_dim]
#         self.eigenvalues = s[:max_dim] ** 2

#     def scores(self, Xtest, max_dim=None):
#         """The method returns the scores, i.e. the coordinates of the
#         points in the array Xtest, in the coordinate system given by
#         the principal axes, of the extreme points (with norm greater
#         than self.threh_predict) rescaled through the mapping
#          x -->  ||x||^{-beta} *  x  where beta = self.beta. The model
#         must have been fit beforehand. If `Xtest` is the training data,
#         this gives the usual PCA scores of the rescaled extreme data.

#         """
#         if self.principal_axes is None:
#             raise ValueError("The model has not been fitted")
#         if max_dim is None:
#             p = np.shape(self.principal_axes)[1]
#         else:
#             p = max_dim

#         if self.thresh_predict is None:
#             self.thresh_predict = self.thresh_train

#         Xt = Xtest
#         norm_Xt = np.sum(Xt**2, axis=1) ** (1 / 2)
#         # define extreme samples
#         is_extreme = norm_Xt > self.thresh_predict
#         X_ext = Xt[is_extreme]
#         R_ext = norm_Xt[is_extreme]
#         ang_ext = X_ext / (R_ext ** (self.beta)).reshape(-1, 1)

#         if not self.centered:
#             Zang = ang_ext
#         else:
#             Zang = ang_ext - self.mean_angle
#         coords_in_pr_axes = Zang @ self.principal_axes[:, :p]
#         mask = is_extreme
#         return coords_in_pr_axes, X_ext, mask

#     def predict(self, X_test, max_dim=None):
#         """
#         projects X_test onto the

#         """
#         if self.principal_axes is None:
#             raise ValueError("The model has not been fitted")
#         if max_dim is None:
#             p = np.shape(self.principal_axes)[1]
#         else:
#             p = max_dim
#         Scores, X_ext, mask = self.scores(X_test, max_dim)
#         recons_angle_intermediate = Scores @ ((self.principal_axes[:, :p]).transpose())
#         if self.centered:
#             recons_angle = recons_angle_intermediate + self.mean_angle
#         else:
#             recons_angle = recons_angle_intermediate
#         R_ext = (np.sum(X_ext**2, axis=1)) ** (1 / 2)
#         recons_X = recons_angle * (R_ext**self.beta).reshape(-1, 1)
#         return recons_X, X_ext, mask

#     def screeplot(self):
#         if self.principal_axes is None:
#             raise ValueError("The model has not been fitted")
#         eigenvals = self.eigenvalues
#         # Create the scree plot
#         plt.figure(figsize=(10, 6))
#         plt.plot(eigenvals, "bo-", markersize=8, label="Eigenvalues")
#         # Add labels and title
#         plt.xlabel("Principal Component Index")
#         plt.ylabel("Eigenvalue")
#         plt.title("Scree Plot")
#         plt.legend()
#         plt.grid(True)
#         plt.show()

#     def recons_error(self, X_test, max_dim, test_beta=None):
#         """The test data are rescaled to ensure existence of second
#         moments, namely each test data is transformed following

#         z   =   || xtest ||^{- beta}  xtest

#         where  beta <= 1  is given by the argument
#         `test_beta`. Default is `test_beta=None`, it which case `test_beta`
#         is set to the training value self.beta which default is 1.

#         """
#         if test_beta is None:
#             beta = self.beta
#         else:
#             beta = test_beta
#         recons_X, X_ext, mask = self.predict(X_test, max_dim)

#         X_ext_norm = np.sum(X_ext**2, axis=1) ** (1 / 2)
#         Z = X_ext / (X_ext_norm**beta).reshape(-1, 1)
#         recons_Z = recons_X / (X_ext_norm**beta).reshape(-1, 1)
#         Err = np.sum((Z - recons_Z) ** 2, axis=1)
#         return Err



from .dataset_generation import (
    gen_rv_dirimix,
    gen_classif_data_diriClasses,
    gen_target_CovariateRV,
    gen_rv_functional_data,
    gen_rv_functional_data_gaussianNoise,
    gen_subfaces,
    normalize_param_dirimix,
    gen_dirichlet,
    pdf_dirichlet,
    pdf_dirimix,
    gen_dirimix,
    plot_pdf_dirimix_2D,
    plot_pdf_dirimix_3D,
    # gen_PositiveStable,
    gen_multilog,
    transform_target_lin,
    inv_transform_target_lin,
    transform_target_nonlin,
    inv_transform_target_nonlin
 
)

from .EVT_basics import (
    round_signif,
    hill_estimator,
    rank_transform,
    rank_transform_test,
    test_indep_radius_rest,
    plot_indep_radius_rest
)



"""
Plotting utilities for the `supervised` module.
"""

import matplotlib.pyplot as plt

def plot_classif(X, y_test, y_pred, title=None):
    """
    Display points classified according to predictions and actual values.
    The covariate points are projected onto their first and last components.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Input samples for plotting.

    y_test : array-like, shape (n_samples,)
        True labels for the input samples.

    y_pred : array-like, shape (n_samples,)
        Predicted labels for the input samples.

    title : str, optional
        Title of the plot. Default is "Classification Results".
    """
    plt.figure(figsize=(10, 6))
    plt.scatter(
        X[:, 0][(y_pred == 0) & (y_test == 0)],
        X[:, -1][(y_pred == 0) & (y_test == 0)],
        color="blue",
        marker="o",
        label="True Negative",
        alpha=0.5,
    )
    plt.scatter(
        X[:, 0][(y_pred == 0) & (y_test == 1)],
        X[:, -1][(y_pred == 0) & (y_test == 1)],
        color="red",
        marker="x",
        label="False Negative",
        alpha=0.5,
    )
    plt.scatter(
        X[:, 0][(y_pred == 1) & (y_test == 0)],
        X[:, -1][(y_pred == 1) & (y_test == 0)],
        color="blue",
        marker="x",
        label="False Positive",
        alpha=0.5,
    )
    plt.scatter(
        X[:, 0][(y_pred == 1) & (y_test == 1)],
        X[:, -1][(y_pred == 1) & (y_test == 1)],
        color="red",
        marker="o",
        label="True Positive",
        alpha=0.5,
    )
    if title is None:
        title = "Classification Results"
    plt.title(title)
    plt.xlabel("Feature 1")
    plt.ylabel("Feature p")
    plt.legend()
    plt.show()

def plot_predictions(y_true, y_pred, title=None):
    """
    Display predicted vs actual values.

    Parameters
    ----------
    y_true : array-like, shape (n_samples,)
        True values for comparison.

    y_pred : array-like, shape (n_samples,)
        Predicted values for comparison.

    title : str, optional
        Title of the plot. Default is "True vs Predicted Values".
    """
    plt.figure(figsize=(10, 6))
    plt.scatter(y_true, y_pred, color="blue", marker="o")
    plt.plot([y_true.min(), y_true.max()], [y_true.min(), y_true.max()], "r--")
    plt.xlabel("True Values")
    plt.ylabel("Predictions")
    if title is None:
        title = "True vs Predicted Values"
    plt.title(title)
    plt.show()
# """
# Plotting utilities for the `supervised` module
# """

# import matplotlib.pyplot as plt


# def plot_classif(X, y_test, y_pred, title=None):
#     """
#     Display points classified according to predictions and actual values.
#     the covariate points are projected onto there first and last component.

#     Parameters
#     ----------
#     X : array-like of shape (n_samples, n_features)
#         The input samples.

#     y_test : array-like of shape (n_samples,)
#         The true labels.

#     y_pred : array-like of shape (n_samples,)
#         The predicted labels.
#     """
#     plt.figure(figsize=(10, 6))
#     plt.scatter(
#         X[:, 0][(y_pred == 0) & (y_test == 0)],
#         X[:, -1][(y_pred == 0) & (y_test == 0)],
#         color="blue",
#         marker="o",
#         label="True Negative",
#         alpha=0.5,
#     )
#     plt.scatter(
#         X[:, 0][(y_pred == 0) & (y_test == 1)],
#         X[:, -1][(y_pred == 0) & (y_test == 1)],
#         color="red",
#         marker="x",
#         label="False Negative",
#         alpha=0.5,
#     )
#     plt.scatter(
#         X[:, 0][(y_pred == 1) & (y_test == 0)],
#         X[:, -1][(y_pred == 1) & (y_test == 0)],
#         color="blue",
#         marker="x",
#         label="False Positive",
#         alpha=0.5,
#     )
#     plt.scatter(
#         X[:, 0][(y_pred == 1) & (y_test == 1)],
#         X[:, -1][(y_pred == 1) & (y_test == 1)],
#         color="red",
#         marker="o",
#         label="True Positive",
#         alpha=0.5,
#     )
#     if title is None:
#         title = "Classification Results"
#     plt.title(title)
#     plt.xlabel("Feature 1")
#     plt.ylabel("Feature p")
#     plt.legend()
#     plt.show()


# def plot_predictions(y_true, y_pred, title=None):
#     """
#     Display predicted vs actual values.

#     Parameters

#     y_true : array-like of shape (n_samples,)
#         The true values.

#     y_pred : array-like of shape (n_samples,)
#         The predicted values.
#     """
#     plt.figure(figsize=(10, 6))
#     plt.scatter(y_true, y_pred, color="blue", marker="o")
#     plt.plot([y_true.min(), y_true.max()], [y_true.min(), y_true.max()], "r--")
#     plt.xlabel("True Values")
#     plt.ylabel("Predictions")
#     if title is None:
#         title = "True vs Predicted Values"
#     plt.title(title)
#     plt.show()



"""
File: mvsets.py
Author: Anne Sabourin
Date: 2025


"""
import numpy as np
from math import isclose
import matplotlib.pyplot as plt
from copy import deepcopy
from MLExtreme.utils.EVT_basics import rank_transform, round_signif



class Angular_bin:
    """Building blocks of angular minimum-volume sets (mvsets)
    constructed from a regular paving of the unit sphere for the
    infinite norm. Geometrically, each grid is a rectangle on the
    positive orthant of the unit sphere defined from the infinite
    norm, of the kind:

    prod_{i in {1, ..., d}, i != i_0} [j_i/J, (j_i+1)/J) x {e_{i_0}}

    where j_i in {0, ..., J-1}, J is the chosen number of basic
    subdivisions of the interval [0, 1) and i_0 is the index of the
    canonical basis vector e_{i_0} normal to the face where the bin is
    located.

    Internally a bin is encoded as a np.array of integers in 0, ..., J-1
    of dimension d, containing the lower end points of the intervals in
    the product, except for the i_0^{th} entry which is equal to J. A hash
    function is provided to enable quick responses to queries about the
    existence of keys in a dictionary of `angular_bin` class instances.
    """

    def __init__(self, coords: np.array):
        ##self.subface = np.argmax(coords)
        discrete_coord = np.floor(coords).astype(int)
        J = np.max(discrete_coord)
        id_max = np.where(discrete_coord == J)[0]
        if len(id_max) > 1:
            raise ValueError("A bin cannot be on an edge.")
        self.coords = discrete_coord

    def __hash__(self):
        coords = np.array(self.coords)
        J = np.max(coords)
        dim = len(coords)
        exponents = (J + 1) ** np.arange(dim)
        return int(np.sum(exponents * coords))

    def __eq__(self, other):
        if isinstance(other, Angular_bin):
            return np.array_equal(self.coords, other.coords)
        return False


def get_angular_bin(point: np.array, J: int) -> Angular_bin:
    """Get the angular bin for a given point.

    Parameters:
    -----------
    point : np.array
        The point for which to find the angular bin.
    J : int
        The number of subdivisions of the interval [0, 1).

    Returns:
    --------
    Angular_bin
        The angular bin corresponding to the given point.
    """
    angle = point / np.max(point)
    edge_indices = np.where(np.isclose(angle, 1))[0]
    # move the point away from the edge by decreasing all but the
    # one random coordinate by a small enough value so that the shifted
    # point belongs to the next adjacent bin.
    if len(edge_indices) > 1:
        j = np.random.randint(0, len(edge_indices))
        other_indices = np.delete(edge_indices, j)
        for i in other_indices:
            angle[i] = angle[i] - 1 / (2 * J)

    return Angular_bin(J * angle)


def sort_bins_by_mass(X: np.array, J: int) -> list[Angular_bin]:
    """Sort angular bins by their mass.

    Parameters:
    -----------
    X : np.array
        The input data.
    J : int
        The number of subdivisions of the interval [0, 1).

    Returns:
    --------
    list[Angular_bin]
        A list of angular bins sorted by mass in descending order.
    """
    bin_dict = {}
    n = np.shape(X)[0]
    # pdb.set_trace()
    for i in range(n):
        x = X[i,]
        ang_bin = get_angular_bin(x, J)
        if ang_bin not in bin_dict:
            bin_dict[ang_bin] = 1
        else:
            bin_dict[ang_bin] += 1

    sorted_list = sorted(bin_dict.items(), key=lambda x: x[1], reverse=True)
    return sorted_list


class Xmvset:
    """
    Minimum-volume set estimation and anomaly scoring.

    Implements the methods developed in [1].

    Compared with the implementation proposed in [1] the only  difference resides on the data structure manipulation performs to affect points to the building blocks of the mass-volume sets. This doess not affect the output.

    References
    _________
    [1] Thomas, A., Clémençon, S., Gramfort, A., & Sabourin, A. (2017, April). Anomaly Detection in Extreme Regions via Empirical MV-sets on the Sphere. In AISTATS (Vol. 54).
    """

    def __init__(self, thresh_train=None, thresh_predict=None, J=None):
        """Initialize the Xmvset model.

        Parameters:
        -----------
        thresh_train : float, optional
            Threshold for identifying extreme points during training.
        thresh_predict : float, optional
            Threshold for identifying extreme points during prediction.
        J : int, optional
            Number of subdivisions of the interval [0, 1).
        """
        self.thresh_train = thresh_train
        self.thresh_predict = thresh_predict
        self.J = J
        self.masses = None
        self.coords = None
        self.Xtrain = None

    def set_default_J(self, X):
        """Set the default value for J based on the input data.

        Parameters:
        -----------
        X : np.array
            The input data.
        """
        n = len(X)
        d = np.shape(X)[1]
        default_J = np.ceil((n / d**2) ** (1 / (2 * (d - 1))))
        self.J = default_J

    def fit(self, X):
        """Fit the model to the input data.

        Parameters:
        -----------
        X : np.array
            The input data.
        """
        # # Standardize X if needed
        # if standardize:
        #     Xt = rank_transform(X)
        # else:
        #     Xt = X
        Xt = X
        norm_Xt = np.max(Xt, axis=1)

        # Set the training threshold if not provided
        if self.thresh_train is None:
            self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

        # define extreme samples
        X_ext = Xt[norm_Xt > self.thresh_train]
        # Update total mass accordingly
        self.number_extremes = len(X_ext)
        # self.total_mass = self.thresh_train * self.number_extremes / len(norm_Xt)

        # set J if not provided
        if self.J is None:
            self.set_default_J(X_ext)
        angular_bins = sort_bins_by_mass(X_ext, self.J)

        masses = np.array([item[1] for item in angular_bins])
        coords = np.array([item[0].coords for item in angular_bins])
        # pdb.set_trace()
        self.masses = masses
        self.coords = coords
        # self.list_angular_bins = angular_bins
        # self.Xtrain = X

    def predict(self, X_test, alpha):  # , standardize=True):
        """Predicts whether extreme points belong to an extreme angular minimum-volume
        set at level alpha.

        Parameters:
        -----------
        X_test : np.array
            Array of test data.
        alpha : float
            Proportion of extreme points contained in the angular minimum-volume set.
        standardize : bool, optional
            If True, the test points are rank-transformed based on the training data.
            Default is True.

        Returns:
        --------
        np.array
            An array of boolean values of the same length as X_test, indicating
            whether each point belongs to the extreme angular minimum-volume set.
        """
        if self.masses is None:
            raise ValueError("the model has not been trained")
        if self.thresh_predict is None:
            self.thresh_predict = self.thresh_train
        # cumulative masses of bins preliminarily sorted in decreasing order
        cum_rel_masses = np.cumsum(self.masses) / self.number_extremes
        # indices of bins where cumulative mass exceeds alpha
        id_sets_enough_mass = np.where(cum_rel_masses >= alpha)[0]
        # first index such that the cumulative mass exceeds alpha
        id_mvset = id_sets_enough_mass[0]
        # create a set of bins that form the mvset at level alpha.
        # These are the bins with index less than id_mvset in the array of bins
        # represented by self.coords
        set_bins_mvset = {Angular_bin(self.coords[i]) for i in range(id_mvset + 1)}
        # pdb.set_trace()
        # select extreme samples
        Norm_Xtest = np.max(X_test, axis=1)
        mask_test = Norm_Xtest >= self.thresh_predict
        X_test_extreme = X_test[mask_test]
        n_extremes_test = len(X_test_extreme)
        # among extreme samples, test which ones belong to the mvset :
        # a sample belongs to a mvset iff the angular component of the sample
        # belongs to a bin  which is part of  the mvset.
        is_not_member = np.zeros(n_extremes_test)
        for i in range(n_extremes_test):
            is_not_member[i] = (
                get_angular_bin(X_test_extreme[i], self.J) not in set_bins_mvset
            )
        return is_not_member, X_test_extreme, mask_test

    def score(self, X_test, angular_only=False):
        """Return the angular score of extreme points defined simply as an
        un-normalized histogram-based estimated angular density.
        The lower, the more abnormal.

        Parameters:
        -----------
        X_test : np.array
            Array of test data.
        angular_only : bool, optional
            If True, only the angular score is returned. Default is False.

        Returns:
        --------
        np.array
            An array of scores of the same length as X_test.
        """
        if self.masses is None:
            raise ValueError("the model has not been trained")
        if self.thresh_predict is None:
            self.thresh_predict = self.thresh_train

        # create a dictionary (angular_bins, num_points)
        bin_dict = {
            Angular_bin(self.coords[i]): self.masses[i] for i in range(len(self.masses))
        }

        # select extreme samples
        Norm_Xtest = np.max(X_test, axis=1)
        mask_test = Norm_Xtest >= self.thresh_predict
        X_test_extreme = X_test[mask_test]
        n_extremes_test = len(X_test_extreme)

        # compute the angular score
        score = np.zeros(n_extremes_test)
        for i in range(n_extremes_test):
            bin_key = get_angular_bin(X_test_extreme[i], self.J)
            if bin_key in bin_dict:
                ang_score = bin_dict[bin_key]  # number of extreme angles in that bin
                if angular_only:
                    score[i] = ang_score
                else:
                    score[i] = ang_score / np.max(X_test_extreme[i]) ** 2
        return score, X_test_extreme, mask_test


# workflow:

# input: data
# - select extremes and make angles/radii

# - make sorted angular histogram
# - make angular mvsets with of defined (relative) mass level
#          --> OUTPUT: ordered list of cases +
#           ---> indicator function(level, point)
#           --> angular extremal scoring function
#           ---> extremal scoring


# """
# File: mvsets.py
# Author: Anne Sabourin
# Date: 2025-05-02
# """

# import numpy as np
# from math import isclose
# import matplotlib.pyplot as plt
# from copy import deepcopy
# from ...utils.EVT_basics import rank_transform, round_signif


# class Angular_bin:
#     """Building blocks of angular minimum-volume sets (mvsets)
#     constructed from a regular paving of the unit sphere for the
#     infinite norm. Geometrically, each grid is a rectangle on the
#     positive orthant of the unit sphere defined from the infinite
#     norm, of the kind:

#     prod_{i in {1, ..., d}, i != i_0} [j_i/J, (j_i+1)/J) x {e_{i_0}}

#     where j_i in {0, ..., J-1}, J is the chosen number of basic
#     subdivisions of the interval [0, 1) and i_0 is the index of the
#     canonical basis vector e_{i_0} normal to the face where the bin is
#     located.

#     Internally a bin is encoded as a np.array of integers in 0, ..., J-1
#     of dimension d, containing the lower end points of the intervals in
#     the product, except for the i_0^{th} entry which is equal to J. A hash
#     function is provided to enable quick responses to queries about the
#         existence of keys in a dictionary of `angular_bin` class instances.

#     """

#     def __init__(self, coords: np.array):
#         ##self.subface = np.argmax(coords)
#         discrete_coord = np.floor(coords).astype(int)
#         J = np.max(discrete_coord)
#         id_max = np.where(discrete_coord == J)[0]
#         if len(id_max) > 1:
#             raise ValueError("A bin cannot be on an edge.")
#         self.coords = discrete_coord

#     def __hash__(self):
#         coords = np.array(self.coords)
#         J = np.max(coords)
#         dim = len(coords)
#         exponents = (J + 1) ** np.arange(dim)
#         return int(np.sum(exponents * coords))

#     def __eq__(self, other):
#         if isinstance(other, Angular_bin):
#             return np.array_equal(self.coords, other.coords)
#         return False


# def get_angular_bin(point: np.array, J: int) -> Angular_bin:
#     angle = point / np.max(point)
#     edge_indices = np.where(np.isclose(angle, 1))[0]
#     # move the point away from the edge by decreasing all but the
#     # one random coordinate by a small enough value so that the shifted
#     # point belongs to the next adjacent bin.
#     if len(edge_indices) > 1:
#         j = np.random.randint(0, len(edge_indices))
#         other_indices = np.delete(edge_indices, j)
#         for i in other_indices:
#             angle[i] = angle[i] - 1 / (2 * J)

#     return Angular_bin(J * angle)


# def sort_bins_by_mass(X: np.array, J: int) -> list[Angular_bin]:
#     bin_dict = {}
#     n = np.shape(X)[0]
#     # pdb.set_trace()
#     for i in range(n):
#         x = X[i,]
#         ang_bin = get_angular_bin(x, J)
#         if ang_bin not in bin_dict:
#             bin_dict[ang_bin] = 1
#         else:
#             bin_dict[ang_bin] += 1

#     sorted_list = sorted(bin_dict.items(), key=lambda x: x[1], reverse=True)
#     return sorted_list


# class Xmvset:
#     """
#     minimum-volume set estimation and anomaly scoring
#     """

#     def __init__(self, thresh_train=None, thresh_predict=None, J=None):
#         self.thresh_train = thresh_train
#         self.thresh_predict = thresh_predict
#         self.J = J
#         self.masses = None
#         self.coords = None
#         self.Xtrain = None

#     def set_default_J(self, X):
#         n = len(X)
#         d = np.shape(X)[1]
#         default_J = np.ceil((n / d**2) ** (1 / (2 * (d - 1))))
#         self.J = default_J

#     def fit(self, X):
#         # # Standardize X if needed
#         # if standardize:
#         #     Xt = rank_transform(X)
#         # else:
#         #     Xt = X
#         Xt = X
#         norm_Xt = np.max(Xt, axis=1)

#         # Set the training threshold if not provided
#         if self.thresh_train is None:
#             self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

#         # define extreme samples
#         X_ext = Xt[norm_Xt > self.thresh_train]
#         # Update total mass accordingly
#         self.number_extremes = len(X_ext)
#         # self.total_mass = self.thresh_train * self.number_extremes / len(norm_Xt)

#         # set J if not provided
#         if self.J is None:
#             self.set_default_J(X_ext)
#         angular_bins = sort_bins_by_mass(X_ext, self.J)

#         masses = np.array([item[1] for item in angular_bins])
#         coords = np.array([item[0].coords for item in angular_bins])
#         # pdb.set_trace()
#         self.masses = masses
#         self.coords = coords
#         # self.list_angular_bins = angular_bins
#         # self.Xtrain = X

#     def predict(self, X_test, alpha):  # , standardize=True):
#         """
#         Predicts whether extreme points belong to an extreme angular minimum-volume
#         set at level alpha.

#         Parameters:
#         -----------
#         Xtest : np.array
#             Array of test data.
#         alpha : float
#             Proportion of extreme points contained in the angular minimum-volume set.
#         standardize : bool, optional
#             If True, the test points are rank-transformed based on the training data.
#             Default is True.

#         Returns: --------
#         np.array *
#             An array of boolean values of the same length as Xtest, indicating
#             whether each point belongs to the extreme angular minimum-volume set.
#         """
#         if self.masses is None:
#             raise ValueError("the model has not been trained")
#         if self.thresh_predict is None:
#             self.thresh_predict = self.thresh_train
#         # cumulative masses of bins preliminarily sorted in decreasing order
#         cum_rel_masses = np.cumsum(self.masses) / self.number_extremes
#         # indices of bins where cumulative mass exceeds alpha
#         id_sets_enough_mass = np.where(cum_rel_masses >= alpha)[0]
#         # first index such that the cumulative mass exceeds alpha
#         id_mvset = id_sets_enough_mass[0]
#         # create a set of bins that form the mvset at level alpha.
#         # These are the bins with index less than id_mvset in the array of bins
#         # represented by self.coords
#         set_bins_mvset = {Angular_bin(self.coords[i]) for i in range(id_mvset + 1)}
#         # pdb.set_trace()
#         # select extreme samples
#         Norm_Xtest = np.max(X_test, axis=1)
#         mask_test = Norm_Xtest >= self.thresh_predict
#         X_test_extreme = X_test[mask_test]
#         n_extremes_test = len(X_test_extreme)
#         # among extreme samples, test which ones belong to the mvset :
#         # a sample belongs to a mvset iff the angular component of the sample
#         # belongs to a bin  which is part of  the mvset.
#         is_not_member = np.zeros(n_extremes_test)
#         for i in range(n_extremes_test):
#             is_not_member[i] = (
#                 get_angular_bin(X_test_extreme[i], self.J) not in set_bins_mvset
#             )
#         return is_not_member, X_test_extreme, mask_test

#     def score(self, X_test, angular_only=False):
#         """return the angular score of extreme points defined simply as an
#         un-normalized histogram-based estimated angular density.
#         The lower, the more abnormal."""
#         if self.masses is None:
#             raise ValueError("the model has not been trained")
#         if self.thresh_predict is None:
#             self.thresh_predict = self.thresh_train

#         # create a dictionary (angular_bins, num_points)
#         bin_dict = {
#             Angular_bin(self.coords[i]): self.masses[i] for i in range(len(self.masses))
#         }

#         # select extreme samples
#         Norm_Xtest = np.max(X_test, axis=1)
#         mask_test = Norm_Xtest >= self.thresh_predict
#         X_test_extreme = X_test[mask_test]
#         n_extremes_test = len(X_test_extreme)

#         # compute the angular score
#         score = np.zeros(n_extremes_test)
#         for i in range(n_extremes_test):
#             bin_key = get_angular_bin(X_test_extreme[i], self.J)
#             if bin_key in bin_dict:
#                 ang_score = bin_dict[bin_key]  # number of extreme angles in that bin
#                 if angular_only:
#                     score[i] = ang_score
#                 else:
#                     score[i] = ang_score / np.max(X_test_extreme[i]) ** 2
#         return score, X_test_extreme, mask_test




# Author: Anne Sabourin
# Date: 2025
"""
Module for Supervised Learning with Heavy-Tailed Covariates

This module supports supervised binary classification and regression tasks with squared error loss, specifically designed for scenarios involving heavy-tailed (regularly varying) covariates.
"""

import numpy as np
from sklearn.metrics import hamming_loss
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold
from copy import deepcopy


class xcovPredictor:
    """
    A prediction model for extreme value analysis, tailored for classification\
    or regression tasks on the tails of the covariate distribution.

    Parameters
    ----------
    task : str
        The type of prediction task, either 'regression' or 'classification'.
    model : object
        The model used for prediction, must have `.fit` and `.predict` methods.
    norm_func : callable, optional
        Function to compute the norm of input data. Defaults to L2 norm.

    Attributes
    ----------
    task : str
        The type of prediction task.
    model : object
        A deep copy of the model provided during initialization.
    norm_func : callable
        The norm function used for selecting extremes.
    thresh_train : float
        Training threshold, set after fitting.
    ratio_train : float
        Ratio of extreme training samples to total training samples, set after fitting.

    Methods
    -------
    fit(X_train, y_train, k=None, thresh_train=None):
        Fit the model using extreme points from the training data.
    predict(X_test, thresh_predict=None):
        Predict the target from extreme test covariates.
    cross_validate(X, y, thresh_train=None, thresh_predict=None, cv=5, scoring=None, random_state=None):
        Evaluate the generalization risk using cross-validation.
    """

    def __init__(self, task, model, norm_func=None):
        if task not in ["regression", "classification"]:
            raise ValueError("task must be either 'classification' or 'regression'")
        internal_model = deepcopy(model)
        self.task = task
        self.model = internal_model
        self.norm_func = norm_func if norm_func else lambda x: np.linalg.norm(x, axis=1)
        self.thresh_train = None
        self.ratio_train = None

    def fit(self, X_train, y_train, k=None, thresh_train=None):
        """
        Fit the model using extreme points from the training data.

        Parameters
        ----------
        X_train : array-like, shape (n_samples, n_features)
            Training input samples.
        y_train : array-like, shape (n_samples,)
            Target values for training.
        k : int, optional
            Number of extreme samples to use for training.
        thresh_train : float, optional
            Threshold for considering extreme samples during training.

        Returns
        -------
        thresh_train : float
            The training threshold used.
        ratio_train : float
            Ratio of extreme points used for training.
        X_train_extreme : array-like, shape (n_extreme_samples, n_features)
            Extreme covariate points used for training.
        """
        if not callable(self.norm_func):
            raise ValueError("norm_func must be callable")

        if k is not None and thresh_train is not None:
            raise ValueError("k and thresh_train cannot both be set at the same time")

        Norm_X_train = self.norm_func(X_train)

        if k is None and thresh_train is None:
            thresh_train = np.quantile(Norm_X_train,
                                       1 - 1 / np.sqrt(len(Norm_X_train)))
        if thresh_train is None:
            thresh_train = np.quantile(Norm_X_train, 1 - k / len(Norm_X_train))

        id_extreme = Norm_X_train >= thresh_train
        if k is None:
            k = np.sum(id_extreme)

        self.thresh_train = thresh_train
        self.ratio_train = k / len(Norm_X_train)

        X_train_extreme = X_train[id_extreme]
        X_train_extreme_unit = X_train_extreme / Norm_X_train[id_extreme][:, np.newaxis]
        y_train_extreme = y_train[id_extreme]

        self.model.fit(X_train_extreme_unit, y_train_extreme)

        return thresh_train, k / len(Norm_X_train), X_train_extreme

    def predict(self, X_test, thresh_predict=None):
        """
        Predict the target from extreme test covariates.

        Parameters
        ----------
        X_test : array-like, shape (n_samples, n_features)
            Test input samples.
        thresh_predict : float, optional
            Threshold for selecting extreme points during prediction.

        Returns
        -------
        y_pred_extreme : array-like, shape (n_extreme_samples,)
            Predicted labels for extreme points.
        X_test_extreme : array-like, shape (n_extreme_samples, n_features)
            Extreme points from the test data.
        mask_test : array-like, shape (n_samples,)
            Boolean mask indicating extreme points.
        """
        if self.thresh_train is None:
            raise ValueError("Model has not been fitted yet.")

        if thresh_predict is None:
            thresh_predict = self.thresh_train

        Norm_X_test = self.norm_func(X_test)
        mask_test = Norm_X_test >= thresh_predict
        X_test_extreme = X_test[mask_test]

        X_test_extreme_unit = X_test_extreme / Norm_X_test[mask_test][:, np.newaxis]
        y_pred_extreme = self.model.predict(X_test_extreme_unit)

        return y_pred_extreme, X_test_extreme, mask_test

    def cross_validate(
        self,
        X,
        y,
        thresh_train=None,
        thresh_predict=None,
        cv=5,
        scoring=None,
        random_state=None
    ):
        """
        Perform cross-validation and return the mean score, standard deviation, and scores.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input samples.
        y : array-like, shape (n_samples,)
            Target values.
        thresh_train : float, optional
            Threshold for extreme values during training.
        thresh_predict : float, optional
            Threshold for extreme values during prediction.
        cv : int, optional
            Number of folds for cross-validation. Default is 5.
        scoring : callable, optional
            Scoring function to evaluate predictions.
        random_state : int, optional
            Seed for random number generation.

        Returns
        -------
        mean_score : float
            Mean score from cross-validation.
        std_err_mean : float
            Estimated standard deviation of the mean score.
        scores : list of float
            Scores from each fold of the cross-validation.
        """
        scores = []

        if not callable(self.norm_func):
            raise ValueError("norm_func must be callable")

        if scoring is None:
            scoring = (
                hamming_loss if self.task == "classification" else mean_squared_error
            )

        if not callable(scoring):
            raise ValueError("scoring must be callable or None")

        Norm_X = self.norm_func(X)

        if thresh_train is None:
            thresh_train = np.quantile(Norm_X, (1 - 1 / np.sqrt(len(Norm_X))))

        if thresh_predict is None:
            thresh_predict = thresh_train

        id_extreme_train = Norm_X >= thresh_train
        id_extreme_predict = Norm_X >= thresh_predict

        kf = KFold(n_splits=cv, shuffle=True, random_state=random_state)

        for train_index, test_index in kf.split(X):
            size_train_ex = np.sum(id_extreme_train[train_index])
            size_predict_ex = np.sum(id_extreme_predict[test_index])
            if size_train_ex <= 2 or size_predict_ex <= 0:
                continue

            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]

            try:
                self.fit(X_train, y_train, k=None, thresh_train=thresh_train)
            except Exception as e:
                ratio_train = size_train_ex / len(train_index)
                print(
                    f"Warning: An error occurred for extreme ratio (training) = {ratio_train:4f}. Error: {e}"
                )
                continue

            y_pred_extreme, _, mask_test = self.predict(
                X_test, thresh_predict=thresh_predict
            )
            result = scoring(y_test[mask_test], y_pred_extreme)
            scores.append(result)

        mean_size_train = np.sum(id_extreme_train) * (cv - 1) / cv
        mean_size_predict = np.sum(id_extreme_predict) / cv
        order_error = 1 / np.sqrt(mean_size_predict) + 1 / np.sqrt(mean_size_train)

        return np.mean(scores), order_error, scores


# import numpy as np
# from sklearn.metrics import hamming_loss
# from sklearn.metrics import mean_squared_error
# from sklearn.model_selection import KFold
# from copy import deepcopy


# class xcovPredictor:
#     """
#     A prediction model (either classification or regression)
#      designed specifically for making predictions on the tails of the
#      covariate distribution.

#      Parameters
#     ----------

#     task: str The type of prediction task
#      envisioned.  Must be either 'prediction' or 'classification'.

#     model : object The model to be used for prediction. Must have a
#         `.fit` method and a `.predict' method.

#     norm_func : callable, optional A function that computes a norm of
#         the input data. Default is L2 norm.


#     Attributes
#     ----------

#     task : str The type of prediction task
#     envisioned.  Either 'prediction' or 'classification'.

#     model : object The model used for prediction.  It is a
#         `copy.deepcopy` copy of the argument passed to the contructor.

#     norm_func : callable The norm function used at training and
#         testing for selecting extremes.

#     thresh_train: float.  Set to None at initialization and later set
#         to the training threshold after fitting.

#      ratio_train: float, between 0 and 1 Set to None at initialization
#         and later set to the ratio k/n of extreme training sample
#         divided by the training sample size.

#     Methods
#     _________

#     .fit : fit the self.model object using the angle
#     of largest covariates and their associated targets

#     .predict : predicts the target based on new (extreme) covariates

#     .cross_validate : evaluates the generalization risk based on
#     cross-validation."""

#     def __init__(self, task, model, norm_func=None):
#         if task != "regression" and task != "classification":
#             raise ValueError("task must be either 'classification' or 'regression'")
#         internal_model = deepcopy(model)
#         self.task = task
#         self.model = internal_model
#         self.norm_func = norm_func if norm_func else lambda x: np.linalg.norm(x, axis=1)
#         self.thresh_train = None
#         self.ratio_train = None

#     def fit(self, X_train, y_train, k=None, thresh_train=None):
#         """
#         Fit the model using extreme points from the training data where
#         the covariate norm exceeds a high threshold.

#         Parameters
#         ----------
#         X_train : array-like of shape (n_samples, n_features)
#             The training input samples.

#         y_train : array-like of shape (n_samples,)
#             The target values for training.

#         k : int, optional
#             The number of extreme samples used to train the model. Data are
#             ordered according to their norm (`norm_func') and the k largest are
#             selected.

#         thresh_train:  float, optional.
#             The radial threshold above which training samples are considered
#             extreme for training

#         Details
#         ------

#         If `thresh_train=None` and `k=None`, `k` is set to sqrt(n_samples)
#           and thresh_train is set to the empirical 1 - k/n_samples quantile of
#           the norms of the input X_train

#         Setting both thresh_train and k at the same time raises an error.

#         Returns
#         -------
#         thresh_train : float
#             The default threshold for further prediction,
#             set to the training threshold. Also stored in model's
#             attribute .thresh_train

#         ratio_train: float
#             The ratio of extreme points used for training.
#             Also stored in model's attribute .ratio_train

#         X_train : array-like of shape (n_extreme_samples, n_features)
#             The  extreme covariate points actually  used to train the model.
#         """
#         if not callable(self.norm_func):
#             raise ValueError("norm_func must be callable")

#         # check and set k and thresh_train
#         if k is not None and thresh_train is not None:
#             raise ValueError("k and thresh_train cannot both be set at the same time")

#         Norm_X_train = self.norm_func(X_train)

#         if k is None and thresh_train is None:
#             thresh_train = np.quantile(Norm_X_train, 1 - 1 / np.sqrt(len(Norm_X_train)))
#         if thresh_train is None:  ## then k is not none
#             thresh_train = np.quantile(Norm_X_train, 1 - k / len(Norm_X_train))

#         id_extreme = Norm_X_train >= thresh_train
#         if k is None:
#             k = np.sum(id_extreme)
#         # update model with training threshold and k/n ratio
#         self.thresh_train = thresh_train
#         self.ratio_train = k / len(Norm_X_train)

#         X_train_extreme = X_train[id_extreme]
#         X_train_extreme_unit = X_train_extreme / Norm_X_train[id_extreme][:, np.newaxis]
#         y_train_extreme = y_train[id_extreme]

#         self.model.fit(X_train_extreme_unit, y_train_extreme)

#         return thresh_train, k / len(Norm_X_train), X_train_extreme

#     def predict(self, X_test, thresh_predict=None):
#         """
#         Predict the target from  extreme test covariates.

#         Parameters
#         ----------
#         X_test : array-like of shape (n_samples, n_features)
#             The test input (covariate) samples.

#         thresh_predict : float, optional
#             The threshold value to select extreme points. If not provided,
#             the threshold used during fitting will be used by default.


#         Returns
#         -------
#         y_pred : array-like of shape (n_extreme_samples,)
#             The predicted labels for the extreme points.

#         X_test : array-like of shape (n_extreme_samples, n_features)
#             The  extreme points from the test data.

#         mask_test : array-like of shape (n_samples,)
#             A boolean mask indicating which points are extreme.

#         """
#         if self.thresh_train is None:
#             raise ValueError("Model has not been fitted yet.")

#         if thresh_predict is None:
#             thresh_predict = self.thresh_train

#         Norm_X_test = self.norm_func(X_test)
#         mask_test = Norm_X_test >= thresh_predict
#         X_test_extreme = X_test[mask_test]

#         X_test_extreme_unit = X_test_extreme / Norm_X_test[mask_test][:, np.newaxis]
#         y_pred_extreme = self.model.predict(X_test_extreme_unit)

#         return y_pred_extreme, X_test_extreme, mask_test

#     def cross_validate(
#         self,
#         X,
#         y,
#         thresh_train=None,
#         thresh_predict=None,
#         cv=5,
#         scoring=None,
#         random_state=None,
#     ):
#         """Perform cross-validation and return the mean score
#         (defined by the scoring function), the estimated standard
#         deviation of the mean, and the array of scores.

#         Follows mostly Aghbalou et al's CV scheme (K-fold), see [1]
#         below.  Differently from the paper the radial threshold for
#         prediction/test may be chosen different from the radial
#         threshold for training.

#         Parameters
#         ----------
#         X : array-like of shape (n_samples, n_features)
#             The input samples.

#         y : array-like of shape (n_samples,)
#             The target values.

#         thresh_train : float, optional
#             The threshold for considering extreme values during training.
#             By default, set to the 1- 1/sqrt(n) quantile of the covariates norms.

#         thresh_predict : float, optional
#             The threshold for considering extreme values during prediction.
#             By default, set to thresh_train
#         cv : int, optional
#             The number of folds for cross-validation. Default is 5.

#         scoring : callable, optional
#             The scoring function to evaluate predictions.
#             Default is `hamming_loss` (i.e. 0-1 loss) for classification tasks
#             and `mean_squared_error` for regression tasks.

#         random_state : int, optional
#             Seed for the random number generator for shuffling the data.

#         Returns
#         -------
#         mean_score : float
#             The mean score from cross-validation.

#         std_err_mean : float
#             The estimated standard deviation of the mean score.

#         scores : list of float
#             The scores from each fold of the cross-validation.

#         Details
#         ---------
#         see:

#         [1] Aghbalou, A., Bertail, P., Portier, F., & Sabourin, A. (2024).
#         Cross-validation on extreme regions. Extremes, 27(4), 505-555.

#         """
#         scores = []
#         # begin as in the fit method
#         if not callable(self.norm_func):
#             raise ValueError("norm_func must be callable")

#         # check scoring function
#         if scoring is None:
#             if self.task == "classification":
#                 scoring = hamming_loss
#             if self.task == "regression":
#                 scoring = mean_squared_error

#         if not callable(scoring):
#             raise ValueError("scoring must be callable or None")

#         # check and set thresh_train, as in fit method:
#         # Doing so outside the fit method permits to
#         # discard folds where there are too few extremes in the training and
#         # validation set, without fitting the model first and
#         # thus saving computational time.
#         Norm_X = self.norm_func(X)

#         if thresh_train is None:
#             thresh_train = np.quantile(Norm_X, (1 - 1 / np.sqrt(len(Norm_X))))

#         # id_extreme = Norm_X >= thresh_train
#         # k = np.sum(id_extreme)

#         if thresh_predict is None:
#             thresh_predict = thresh_train

#         # which data are considered extreme for training and testing:
#         # logical mask vectors of size len(y)
#         id_extreme_train = Norm_X >= thresh_train
#         id_extreme_predict = Norm_X >= thresh_predict

#         # K-fold train/test indices
#         kf = KFold(n_splits=cv, shuffle=True, random_state=random_state)
#         # CV loop
#         for train_index, test_index in kf.split(X):
#             size_train_ex = np.sum(id_extreme_train[train_index])
#             size_predict_ex = np.sum(id_extreme_predict[test_index])
#             if size_train_ex <= 2 or size_predict_ex <= 0:
#                 continue
#             # Split the data into training and testing sets
#             X_train, X_test = X[train_index], X[test_index]
#             y_train, y_test = y[train_index], y[test_index]

#             # Fit the model on the training data
#             try:
#                 self.fit(X_train, y_train, k=None, thresh_train=thresh_train)

#             except Exception as e:
#                 ratio_train = size_train_ex / len(train_index)
#                 print(
#                     f"Warning: An error occurred for \
#                 extreme ratio (training) = {ratio_train:4f}. Error: {e}"
#                 )
#                 continue
#             # Predict on the testing data
#             y_pred_extreme, _, mask_test = self.predict(
#                 X_test, thresh_predict=thresh_predict
#             )

#             # Calculate the score
#             result = scoring(y_test[mask_test], y_pred_extreme)
#             scores.append(result)
#         # compute order of magnitude of the error following aghbalou's results
#         # using max(thresh_predict,thresh_train) as a conservative threshold

#         mean_size_train = np.sum(id_extreme_train) * (cv - 1) / cv
#         mean_size_predict = np.sum(id_extreme_predict) / cv
#         #        min_size = np.minimum(mean_size_predict, mean_size_train)
#         order_error = 1 / np.sqrt(mean_size_predict) + 1 / np.sqrt(mean_size_train)
#         #        return np.mean(scores), np.std(scores)/np.sqrt(len(scores)), scores
#         return np.mean(scores), order_error, scores



import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import svd

def CV_recons_error(dataset, nexpe=100, p=3, idex=None, ratioTest=0.3, method='proposed'):
    """
    Calculate the cross-validated reconstruction error of extreme angles in a dataset.

    Parameters:
    -----------
    dataset : array-like of shape (n_samples, n_features)
        The input dataset.

    nexpe : int, optional
        Number of experiments to perform. Default is 100.

    p : int, optional
        The dimension of the projection space. Default is 3.

    idex : array-like, optional
        Index set of extreme observations.

    ratioTest : float, optional
        The ratio of observations among extremes to be put aside as a 'test set'. Default is 0.3.

    method : str, optional
        The method to use for reconstruction. Options are 'proposed', 'full', and 'part'. Default is 'proposed'.

    Returns:
    --------
    error : array of shape (nexpe,)
        The reconstruction errors for each experiment.
    """
    if idex is None:
        raise ValueError("idex must be provided and cannot be None.")

    k = len(idex)
    n, d = dataset.shape
    radii = np.sqrt(np.sum(dataset**2, axis=1))
    angles = dataset / radii[:, np.newaxis]

    error = np.zeros(nexpe)

    for iexp in range(nexpe):
        # Shuffle the indices and split into training and test sets
        shuffledIdex = np.random.permutation(idex)
        idexTest = shuffledIdex[:int(np.floor(k * ratioTest))]
        idexTrain = shuffledIdex[int(np.floor(k * ratioTest)):]

        exanglesTrain = angles[idexTrain, :]
        exanglesTest = angles[idexTest, :]

        if method == 'proposed':
            # Perform SVD on the training angles and project the test angles
            svTrain = svd(exanglesTrain, full_matrices=False)
            projectTest = exanglesTest @ svTrain[2][:, :p] @ svTrain[2][:, :p].T
            errorTest = np.sum((exanglesTest - projectTest)**2)
            error[iexp] = errorTest

        elif method == 'full':
            # Use the full dataset minus the test set for training
            idTrainFull = np.setdiff1d(np.arange(n), idexTest)
            anglesTrainFull = angles[idTrainFull, :]
            svTrainFull = svd(anglesTrainFull, full_matrices=False)
            projectTestFull = exanglesTest @ svTrainFull[2][:, :p] @ svTrainFull[2][:, :p].T
            errorTestFull = np.sum((exanglesTest - projectTestFull)**2)
            error[iexp] = errorTestFull

        elif method == 'part':
            # Use a subset of the full dataset for training
            idTrainFull_red = np.random.choice(np.setdiff1d(np.arange(n), idexTest), len(idexTrain), replace=False)
            anglesTrainFull_red = angles[idTrainFull_red, :]
            svTrainFull_red = svd(anglesTrainFull_red, full_matrices=False)
            projectTestFull_red = exanglesTest @ svTrainFull_red[2][:, :p] @ svTrainFull_red[2][:, :p].T
            errorTestFull_red = np.sum((exanglesTest - projectTestFull_red)**2)
            error[iexp] = errorTestFull_red

        else:
            raise ValueError("method must be 'proposed', 'full', or 'part'.")

    return error

def empMomentFourier(data, freq, exratiomax, exratiomin, graph=True, selectK=50):
    """
    Check the weak convergence of the distribution of angles conditional to a large radius.

    Parameters:
    -----------
    data : array-like of shape (n_samples, n_features)
        The input data.

    freq : int
        The frequency of the Fourier function.

    exratiomax : float
        The maximum ratio of extreme values to consider.

    exratiomin : float
        The minimum ratio of extreme values to consider.

    graph : bool, optional
        Whether to plot the results. Default is True.

    selectK : int, optional
        The value of k to highlight in the plot. Default is 50.

    Returns:
    --------
    values : array of shape (2, kmax - kmin + 1)
        The mean and standard deviation of the projections.
    """
    n, d = data.shape

    # Define the Fourier function based on the frequency
    if freq % 2 == 0:
        fourierFun = np.cos(freq * np.arange(1, d + 1) / d * 2 * np.pi) * np.sqrt(2)
    else:
        fourierFun = np.sin(freq * np.arange(1, d + 1) / d * 2 * np.pi) * np.sqrt(2)

    radii = np.sqrt(np.sum(data**2, axis=1))
    permut = np.argsort(radii)[::-1]  # Indices sorted by decreasing radii
    sradii = np.sort(radii)[::-1]  # Sorted radii
    angles = data / radii[:, np.newaxis]  # Scale data by radii

    kmin = int(np.floor(exratiomin * n))
    kmax = int(np.floor(exratiomax * n))

    values = np.zeros((2, kmax - kmin + 1))
    sortedAngles = angles[permut[:kmax], :]
    projections = np.abs(sortedAngles @ fourierFun)  # Absolute value of projection
    means = np.cumsum(projections) / np.arange(1, kmax + 1)
    variances = np.cumsum((projections - means)**2) / np.arange(1, kmax + 1)
    sdev_est = np.sqrt(variances) / np.sqrt(np.arange(1, kmax + 1))

    values[0, :] = means[kmin:kmax]
    values[1, :] = sdev_est[kmin:kmax]

    if graph:
        plt.figure(figsize=(10, 6))
        plt.plot(np.arange(kmin, kmax + 1), values[0, :], label=f"Frequency = {freq}", color='black')
        plt.fill_between(np.arange(kmin, kmax + 1),
                         values[0, :] - 1.64 * values[1, :],
                         values[0, :] + 1.64 * values[1, :], color='blue', alpha=0.3)
        plt.axvline(x=selectK, linestyle='--', color='gray')
        plt.xlabel('k')
        plt.ylabel(r'$E | < \Theta_t, h_j > |$')
        plt.xticks([kmin, selectK // 2, selectK, selectK * 3 // 2, 2 * selectK, 3 * selectK, 4 * selectK, kmax])
        plt.yticks(np.round([np.min(values[0, :]), np.max(values[0, :])], 2))
        plt.grid(True)
        plt.legend()
        plt.show()

    return values
